ContractLeft

Trait ContractLeft 

pub trait ContractLeft {
    // Required method
    fn contract_left(&self, I: Range<usize>, target_len: usize) -> Range<usize>;
}
Contracting a search pattern from the left.

Required Methods

fn contract_left(&self, I: Range<usize>, target_len: usize) -> Range<usize>

Takes the interval I of a pattern cP in an SBWT, where c is a single character, and returns the interval of pattern P. If I is the empty string, then this should return back the interval of the empty string.

Implementors